#!env python
#
# black bench.py test_bench.py
# .direnv/python-3.8.5/bin/flake8 --max-line-length 100 bench.py test_bench.py
#
# sudo .direnv/python-3.7.3/bin/python bench.py --file-count-ro 5 --file-count-rw 1 --file-size 256
#
import abc
import argparse
import asyncio
import asyncpg
import concurrent.futures
import logging
import os
import random
import sh
import struct
import sys
import datetime
import uuid


class IStorage(abc.ABC):
    def __init__(self, args):
        self.args = args

    @abc.abstractmethod
    async def init(self):
        pass

    @abc.abstractmethod
    async def uninit(self):
        pass

    @abc.abstractmethod
    async def image_list(self):
        pass

    @abc.abstractmethod
    async def image_create(self, image):
        pass

    @abc.abstractmethod
    def image_path(self, image):
        pass

    @abc.abstractmethod
    async def image_delete(self, image):
        pass

    @abc.abstractmethod
    async def clobber(self):
        pass


class Pool(IStorage):
    def __init__(self, args):
        super().__init__(args)

    async def init(self):
        self.rbd = sh.sudo.bake("rbd", f"--pool={self.name}")
        self.ceph = sh.sudo.bake("ceph")
        self.image_size = self.args.file_size * 2

    async def uninit(self):
        pass

    async def image_list(self):
        try:
            self.rbd.ls()
        except sh.ErrorReturnCode_2 as e:
            if "No such file or directory" in e.args[0]:
                return []
            else:
                raise
        return [image.strip() for image in self.rbd.ls()]

    def image_path(self, image):
        return f"/dev/rbd/{self.name}/{image}"

    async def image_create(self, image):
        logging.warning(
            f"rdb --pool {self.name} create --size={self.image_size} {image}"
        )
        self.rbd.create(f"--size={self.image_size}", f"--data-pool={self.name}-data", image)
        self.rbd.feature.disable(
            f"{self.name}/{image}", "object-map", "fast-diff", "deep-flatten"
        )
        self.rbd.device("map", image)
        sh.sudo("chmod", "777", self.image_path(image))

    async def image_delete(self, image):
        logging.warning(f"rdb --pool {self.name} remove {image}")
        self.rbd.device.unmap(f"{self.name}/{image}", _ok_code=(0, 22))
        self.rbd.remove(image)

    async def images_clobber(self):
        for image in await self.image_list():
            image = image.strip()
            await self.image_delete(image)

    async def clobber(self):
        await self.images_clobber()
        self.pool_clobber()


class StorageRO(Pool):
    name = "ro"

    def pool_clobber(self):
        logging.warning(f"ceph osd pool delete {self.name}")
        self.ceph.osd.pool.delete(self.name, self.name, "--yes-i-really-really-mean-it")
        data = f"{self.name}-data"
        logging.warning(f"ceph osd pool delete {data}")
        self.ceph.osd.pool.delete(data, data, "--yes-i-really-really-mean-it")

    def pool_create(self):
        data = f"{self.name}-data"
        logging.warning(f"ceph osd pool create {data}")
        self.ceph.osd(
            "erasure-code-profile",
            "set",
            "--force",
            data,
            "k=4",
            "m=2",
            "crush-failure-domain=host",
        )
        self.ceph.osd.pool.create(data, "erasure", data)
        self.ceph.osd.pool.set(data, 'allow_ec_overwrites', 'true')
        logging.warning(f"ceph osd pool create {self.name}")
        self.ceph.osd.pool.create(self.name)


class StorageRW(IStorage):
    def __init__(self, args):
        super().__init__(args)
        # the rw host is the host with the postgresql server
        # see tests.yml
        self.dsn = "postgres://testuser:testpassword@rw"
        self.db = None

    async def init(self):
        self.db = await asyncpg.connect(dsn=f"{self.dsn}/postgres")

    async def uninit(self):
        await self.db.close()

    async def image_list(self):
        return []

    async def image_create(self):
        image = f"i{uuid.uuid4().hex}"
        db = await asyncpg.connect(dsn=f"{self.dsn}/postgres")
        await db.execute(f"CREATE DATABASE {image}")
        await db.close()
        return image

    def image_path(self, image):
        return f"{self.dsn}/{image}"

    async def image_delete(self, name):
        await self.db.execute(f"DROP DATABASE IF EXISTS {name}")

    async def clobber(self):
        pass


class Database(abc.ABC):
    def __init__(self, args, path):
        self.args = args
        self.path = path

    @abc.abstractmethod
    async def open(self, *args, **kwargs):
        pass

    @abc.abstractmethod
    async def close(self):
        pass


class DatabaseRW(Database):
    @abc.abstractmethod
    async def index(self):
        pass

    @abc.abstractmethod
    async def values(self):
        pass

    @abc.abstractmethod
    async def fill(self):
        pass

    @abc.abstractmethod
    async def size(self):
        pass


class DatabasePostgreSQL(DatabaseRW):
    def __init__(self, args, path):
        super().__init__(args, path)
        self.payloads_define()

    async def open(self):
        self.db = await asyncpg.connect(dsn=self.path)
        await self.db.execute(
            "CREATE TABLE IF NOT EXISTS objects(key BYTEA, content BYTEA)"
        )

    async def close(self):
        return await self.db.close()

    async def index(self):
        async with self.db.transaction():
            async for row in self.db.cursor(
                "SELECT key,LENGTH(content) FROM objects ORDER BY key"
            ):
                yield row

    async def values(self):
        async with self.db.transaction():
            async for row in self.db.cursor("SELECT content FROM objects ORDER BY key"):
                yield row

    async def size(self):
        count = await self.db.fetchrow("SELECT COUNT(*) FROM objects")
        return count[0]

    def payloads_define(self):
        self.payloads = [
            b"0" * 3 * 1024,
            b"1" * 3 * 1024,
            b"2" * 3 * 1024,
            b"3" * 3 * 1024,
            b"4" * 3 * 1024,
            b"5" * 10 * 1024,
            b"6" * 13 * 1024,
            b"7" * 16 * 1024,
            b"8" * 70 * 1024,
            b"9" * 80 * 1024,
        ]

    async def fill(self):
        max_size = self.args.file_size * 1024 * 1024
        size = 0
        while size < max_size:
            async with self.db.transaction():
                payload = random.choice(self.payloads)
                size += len(payload)
                await self.db.execute(
                    "INSERT INTO objects VALUES($1, $2);", os.urandom(32), payload
                )
        return True


class DatabaseRO(Database):
    @abc.abstractmethod
    async def write_index(self, key, size):
        pass

    @abc.abstractmethod
    async def write_content(self, content):
        pass

    @abc.abstractmethod
    async def get(self, key):
        pass


class DatabaseSST(DatabaseRO):

    version = 1

    format_header = "!IQ"
    format_index = "!32sQQ"

    def open(self, mode="rb", size=0):
        self.fd = open(self.path, mode)
        if mode == "wb":
            self.fd.write(
                struct.pack(DatabaseSST.format_header, DatabaseSST.version, size)
            )
            self.size = size
            self.data_offset = (
                self.fd.tell() + struct.calcsize(DatabaseSST.format_index) * size
            )
        else:
            self.header_size = struct.calcsize(DatabaseSST.format_header)
            header = self.fd.read(self.header_size)
            version, self.size = struct.unpack(DatabaseSST.format_header, header)
            self.index_size = struct.calcsize(DatabaseSST.format_index)
            self.data_offset = (
                self.header_size + struct.calcsize(DatabaseSST.format_index) * self.size
            )
            assert version == self.version

    def close(self):
        return self.fd.close()

    def write_index(self, key, size):
        self.fd.write(
            struct.pack(DatabaseSST.format_index, key, self.data_offset, size)
        )

    def write_content(self, content):
        self.fd.write(content)
        self.data_offset += len(content)
        assert self.data_offset == self.fd.tell()

    def get(self, key):
        bounds = self.get_bounds(key)
        if bounds is None:
            return None
        return self.get_content(*bounds)

    def get_content(self, offset, size):
        self.fd.seek(offset)
        return self.fd.read(size)

    def get_bounds(self, search_key):
        min = 0
        max = self.size - 1
        while min <= max:
            pos = int((min + max) / 2)
            self.fd.seek(self.header_size + pos * self.index_size)
            index = self.fd.read(self.index_size)
            key, offset, size = struct.unpack(DatabaseSST.format_index, index)
            logging.warning(f"[{min},{max}] {pos} is {key}")
            if search_key == key:
                return offset, size
            if search_key < key:
                max = pos - 1
            else:
                min = pos + 1
        return None


class Packer(object):
    def __init__(self, args):
        self.args = args
        self.DatabaseRW = DatabasePostgreSQL
        self.DatabaseRO = DatabaseSST

    async def init(self):
        self.storage_ro = StorageRO(self.args)
        await self.storage_ro.init()
        self.storage_rw = StorageRW(self.args)
        await self.storage_rw.init()

    async def uninit(self):
        await self.storage_ro.uninit()
        await self.storage_rw.uninit()

    async def pack(self, image):
        logging.warning(f"Packer.pack {image}")
        db_rw = self.DatabaseRW(self.args, self.storage_rw.image_path(image))
        await self.storage_ro.image_create(image)
        db_ro = self.DatabaseRO(self.args, self.storage_ro.image_path(image))
        await db_rw.open()
        db_ro.open(mode="wb", size=await db_rw.size())
        count = 0
        total_size = 0
        async for key, size in db_rw.index():
            db_ro.write_index(key, size)
            count += 1
            total_size += size
        async for content, in db_rw.values():
            db_ro.write_content(content)
        db_ro.close()
        await db_rw.close()
        await self.storage_rw.image_delete(image)
        return count, total_size


class Bench(object):
    def __init__(self, args):
        self.args = args
        self.stats = {}
        self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=5)
        self.loop = asyncio.get_running_loop()
        #        self.loop.set_debug(True)

        def exception_handler(_, context):
            logging.warning("=========== Bench.exception")
            logging.warning(context)

        self.loop.set_exception_handler(exception_handler)

    async def init(self):
        self.ro = StorageRO(self.args)
        await self.ro.init()
        self.ro.pool_create()
        self.rw = StorageRW(self.args)
        await self.rw.init()

    async def uninit(self):
        await self.ro.uninit()
        await self.rw.uninit()

    async def clobber(self):
        await self.ro.clobber()
        await self.rw.clobber()

    async def reset(self):
        self.ro = StorageRO(self.args)
        await self.ro.init()
        await self.ro.clobber()
        self.rw = StorageRW(self.args)
        await self.rw.init()
        await self.rw.clobber()

    async def writer(self):
        image = await self.rw.image_create()
        logging.warning(f"Bench.writer: {image}")
        d = DatabasePostgreSQL(self.args, self.rw.image_path(image))
        await d.open()
        await d.fill()
        await d.close()
        return image

    async def rw_loop(self, queue):
        logging.warning("Bench.rw_loop: running")

        def create_writer():
            return asyncio.create_task(self.writer())

        self.writers = set()
        for _ in range(self.args.file_count_rw):
            self.writers.add(create_writer())

        while True:
            logging.warning(f"Bench.rw_loop: waiting for {len(self.writers)} writers")
            current = self.writers
            done, pending = await asyncio.wait(
                current, return_when=asyncio.FIRST_COMPLETED
            )
            self.writers = pending
            for task in done:
                result = task.result()
                logging.warning(f"Bench.rw_loop: writer complete {result}")
                await queue.put(result)
                self.writers.add(create_writer())

    async def ro_loop(self, queue):
        while await queue.get():
            queue.task_done()
            images = await self.ro.image_list()
            logging.warning(f"Bench.ro_loop: there are {len(images)} images")
            if len(images) >= self.args.file_count_ro:
                logging.warning("Bench.ro_loop: done")
                return

    async def packer(self, ro_queue, rw_queue):
        self.stats['object_write_count'] = 0
        self.stats['bytes_write'] = 0
        self.packing = set()
        while True:
            image = await rw_queue.get()
            logging.warning(f"Bench.packer: {image}")

            async def pack():
                packer = Packer(self.args)
                await packer.init()
                count, size = await packer.pack(image)
                await packer.uninit()
                return count, size

            t = self.loop.run_in_executor(self.executor, asyncio.run, pack())
            logging.warning(f"Bench.packer: task to run {t}")
            self.packing.add(t)
            done, self.packing = await asyncio.wait(self.packing)
            for t in done:
                logging.warning(f"Bench.packer: result {t.result()}")
                count, size = t.result()
                self.stats['object_write_count'] += count
                self.stats['bytes_write'] += size
                rw_queue.task_done()
                await ro_queue.put(True)

    async def reader(self):
        while True:
            asyncio.sleep(1)
            images = await self.ro.image_list()
            logging.warning(f"Bench.ro_loop: there are {len(images)} images")
            if len(images) >= self.args.file_count_ro:
                logging.warning("Bench.ro_loop: done")
                return

    async def main(self):
        self.stats['begin'] = datetime.datetime.now()
        rw_queue = asyncio.Queue(maxsize=self.args.file_count_rw)
        rw = asyncio.create_task(self.rw_loop(rw_queue))
        ro_queue = asyncio.Queue(maxsize=self.args.file_count_ro)
        ro = asyncio.create_task(self.ro_loop(ro_queue))
        packer = asyncio.create_task(self.packer(ro_queue, rw_queue))
        logging.warning("Bench.main: ro_loop running")
        await ro
        logging.warning("Bench.main: ro_loop complete")
        packer.cancel()
        rw.cancel()
        logging.warning("Bench.main: waiting for packer and rw to complete")
        await asyncio.gather(packer, rw, return_exceptions=True)
        logging.warning("Bench.main: waiting for writers to complete")
        await asyncio.gather(*self.writers, return_exceptions=True)
        logging.warning("Bench.main: waiting for packing to complete")
        await asyncio.gather(*self.packing, return_exceptions=True)
        logging.warning("Bench.main: finished")
        self.stats['end'] = datetime.datetime.now()
        return 0

    def stats_calculate(self):
        def human(size, unit):
            if size < 1024:
                return f'{int(size)}/{unit}s'
            elif size / 1024 < 1024:
                return f'{round(size/1024, 1)}K{unit}/s'
            elif size / (1024 * 1024) < 1024:
                return f'{round(size / (1024 * 1024), 1)}M{unit}/s'
            elif size / (1024 * 1024 * 1024) < 1024:
                return f'{round(size / (1024 * 1024 * 1024), 1)}G{unit}/s'
        elapsed = self.stats['end'] - self.stats['begin']
        seconds = elapsed.total_seconds()
        self.stats['objects_write_per_second'] = human(self.stats['object_write_count'] / seconds, '')
        self.stats['bytes_write_per_second'] = human(self.stats['bytes_write'] / seconds, 'B')

    def stats_print(self):
        self.stats_calculate()
        logging.warning(f"Objects write {self.stats['objects_write_per_second']}")
        logging.warning(f"Bytes write {self.stats['bytes_write_per_second']}")
        return True


class Application(object):
    def parse_args(self, argv):
        parser = argparse.ArgumentParser(description="bench")
        parser.add_argument("--file-count-ro", required=True, type=int)
        parser.add_argument("--file-count-rw", required=True, type=int)
        parser.add_argument("--file-size", required=True, type=int)
        parser.add_argument("--clobber", action="store_true")
        return parser.parse_args(argv)

    async def main(self, argv):
        args = self.parse_args(argv)
        bench = Bench(args)
        await bench.reset()
        await bench.init()
        await bench.main()
        await bench.uninit()
        bench.stats_print()


if __name__ == "__main__":
    sys.exit(asyncio.run(Application().main(sys.argv[1:])))
