#
# coverage run $(which pytest) test_bench.py ; coverage report --include bench.py --show-missing
#
import argparse
import asyncio
import datetime
import bench
import pytest


@pytest.mark.asyncio
async def test_DatabasePostegreSQL(tmpdir):
    args = argparse.Namespace(file_size=0.1)
    s = bench.StorageRW(args)
    await s.init()
    image = await s.image_create()
    d = bench.DatabasePostgreSQL(args, s.image_path(image))
    await d.open()
    await d.fill()
    assert await d.size() > 1
    async for key, size in d.index():
        assert len(key) == 32
        assert size > 1024
    async for content, in d.values():
        assert len(content) > 1024
    await d.close()
    await s.image_delete(image)
    await s.uninit()


def test_DatabaseSST(tmpdir):
    path = f"{tmpdir}/database.sst"
    d = bench.DatabaseSST(None, path)
    size = 1
    key = b"K" * 32
    content = b"something"
    d.open(mode="wb", size=size)
    d.write_index(key, len(content))
    d.write_content(content)

    d = bench.DatabaseSST(None, path)
    d.open()
    assert d.get(key) == content


@pytest.mark.asyncio
async def test_StorageRO():
    args = argparse.Namespace(file_size=256)
    name = "IMAGE"
    s = bench.StorageRO(args)
    await s.init()
    await s.clobber()
    s.pool_create()
    await s.image_create(name)
    p = s.image_path(name)
    assert p.endswith(name)
    something = "SOMETHING"
    open(p, "w").write(something)
    assert open(p).read(len(something)) == something
    assert await s.image_list() == [name]
    await s.images_clobber()
    assert await s.image_list() == []
    await s.clobber()
    assert await s.image_list() == []


@pytest.mark.asyncio
async def test_Packer(tmpdir):
    args = argparse.Namespace(file_size=1)
    storage_rw = bench.StorageRW(args)
    await storage_rw.init()
    image = await storage_rw.image_create()
    d = bench.DatabasePostgreSQL(args, storage_rw.image_path(image))
    await d.open()
    await d.fill()
    assert await d.size() > 1
    await d.close()

    storage_ro = bench.StorageRO(args)
    await storage_ro.init()
    await storage_ro.clobber()
    storage_ro.pool_create()

    packer = bench.Packer(args)
    await packer.init()
    await packer.pack(image)
    await storage_rw.uninit()
    await storage_ro.clobber()
    await storage_ro.uninit()


@pytest.mark.asyncio
async def test_Bench_writer():

    args = argparse.Namespace(
        file_count_rw=1, file_count_ro=1, clobber=False, file_size=1
    )
    b = bench.Bench(args)
    await b.init()
    image = await b.writer()

    d = bench.DatabasePostgreSQL(args, b.rw.image_path(image))
    await d.open()
    assert await d.size() > 0
    await d.close()

    await b.rw.image_delete(image)
    await b.uninit()


@pytest.mark.asyncio
async def test_Bench_rw_loop():

    file_count_rw = 2
    args = argparse.Namespace(
        file_count_rw=file_count_rw, file_count_ro=1, clobber=False, file_size=1
    )
    b = bench.Bench(args)
    await b.init()

    rw_queue = asyncio.Queue(maxsize=file_count_rw)
    rw = asyncio.create_task(b.rw_loop(rw_queue))

    # keep the first image
    image = await rw_queue.get()
    rw_queue.task_done()
    # discard the second image
    await rw_queue.get()
    rw_queue.task_done()

    await rw_queue.join()

    d = bench.DatabasePostgreSQL(args, b.rw.image_path(image))
    await d.open()
    assert await d.size() > 0
    await d.close()

    rw.cancel()

    for task in asyncio.all_tasks():
        if task == asyncio.current_task():
            continue
        try:
            await task
        except asyncio.CancelledError:
            pass
    await b.rw.image_delete(image)

    await b.clobber()
    await b.uninit()


@pytest.mark.asyncio
async def test_Bench_ro_loop():

    args = argparse.Namespace(
        file_count_rw=1, file_count_ro=1, clobber=False, file_size=1
    )

    storage_ro = bench.StorageRO(args)
    await storage_ro.init()
    await storage_ro.clobber()
    storage_ro.pool_create()
    name = "IMAGE"
    await storage_ro.image_create(name)

    b = bench.Bench(args)
    await b.init()

    ro_queue = asyncio.Queue(maxsize=1)
    ro = asyncio.create_task(b.ro_loop(ro_queue))
    await ro_queue.put(True)
    await ro_queue.join()
    await ro

    await storage_ro.clobber()
    await storage_ro.uninit()


@pytest.mark.asyncio
async def test_Bench_packer():

    count_ro = 1
    args = argparse.Namespace(file_count_rw=1, file_count_ro=count_ro, file_size=1)
    b = bench.Bench(args)
    await b.init()
    await b.clobber()
    await b.init()

    image = await b.rw.image_create()
    d = bench.DatabasePostgreSQL(args, b.rw.image_path(image))
    await d.open()
    await d.fill()
    assert await d.size() > 1
    await d.close()

    rw_queue = asyncio.Queue()
    await rw_queue.put(image)
    ro_queue = asyncio.Queue()
    packer = asyncio.create_task(b.packer(ro_queue, rw_queue))
    await ro_queue.get()
    assert len(await b.ro.image_list()) == count_ro
    packer.cancel()


@pytest.mark.asyncio
async def test_Bench_main():

    count_ro = 1
    count_rw = 1
    args = argparse.Namespace(
        file_count_rw=count_rw, file_count_ro=count_ro, clobber=False, file_size=1
    )
    b = bench.Bench(args)
    await b.reset()
    await b.init()
    await b.main()
    actual_count_ro = len(await b.ro.image_list())
    await b.uninit()
    #
    # There are at leaest count_ro images and at most count_ro +
    # count_rw images in the end. The bench stops when there is at
    # least count_ro images but there may be up to count_rw writers in
    # flight and if they all complete there will be count_rw images
    # more when main completes.
    #
    assert actual_count_ro >= count_ro
    assert actual_count_ro <= count_ro + count_rw


@pytest.mark.asyncio
async def test_stats():
    args = argparse.Namespace(file_count_rw=1, file_count_ro=1, file_size=1)
    b = bench.Bench(args)
    now = datetime.datetime.now()
    duration = 60
    before = now - datetime.timedelta(seconds=duration)
    giga = 1024*1024*1024
    b.stats = {
        'begin': before,
        'end': now,
        'object_write_count': 600,
        'bytes_write': 60 * giga,
    }
    b.stats_calculate()
    assert b.stats['objects_write_per_second'] == '10/s'
    assert b.stats['bytes_write_per_second'] == '1.0GB/s'
    assert b.stats_print() is True
